package org.ibex.xt;

import java.io.StringReader;
import java.io.Writer;
import java.io.OutputStream;
import java.io.IOException;
import java.util.*;

import org.ibex.util.*;
import org.ibex.js.JS;
import org.ibex.js.JSScope;

public class JSElement extends JSScope implements XML.Element {
    protected XML.Element wrapped;

    /** Creates an Element around <tt>wrapped</tt>, replacing
     *  references to it in its parent and children with this object. */
    public JSElement(XML.Element wrapped) {
        super(findScope(wrapped));
        this.wrapped = wrapped;

        // remap parent and children
        if (wrapped.getParent() != null) {
            List c = wrapped.getParent().getChildren();
            c.set(c.indexOf(wrapped), this);
        }
        List c = wrapped.getChildren();
        for (int i=0; i < c.size(); i++) ((Tree.Leaf)c.get(i)).setParent(this);
    }

    public void out(OutputStream o) throws IOException { wrapped.out(o); }
    public void out(Writer w) throws IOException { wrapped.out(w); }

    /** Load the attributes into the js scope. */
    protected void loadAttr() {
        try {
            XML.Attributes a = getAttributes();
            for(int i=0; i < a.attrSize(); i++) {
                if (!"http://xt.ibex.org/".equals(a.getUri(i))) continue;
                declare(a.getKey(i));
                put(a.getKey(i), eval(a.getVal(i)));
            }
        } catch (Exception e) { throw new RuntimeException(e); }
    }

    private Object eval(String s) {
        if (s == null) return null;
        StringBuffer ret = new StringBuffer();
        while (s.indexOf("${") != -1) {
            ret.append(s.substring(0, s.indexOf("${")));
            String s2 = s.substring(s.indexOf("${")+2);
            Object app = exec("return (" + s2.substring(0, s2.indexOf('}')) + ");\n");
            s = s.substring(s.indexOf('}') + 1);

            if (!(app == null ||
                  app instanceof String ||
                  app instanceof Number ||
                  app instanceof Boolean))
                throw new RuntimeException("javascripts within ${...} can only return " +
                                           "strings, numbers, and booleans; not a " +
                                           app.getClass().getName());

            ret.append(app == null ? "null" : app.toString());
        }
        ret.append(s);
        return ret.toString();
    }

    public Object exec(String s) {
        try {
            return JS.eval(JS.cloneWithNewParentScope(
                           JS.fromReader("input", 0, new StringReader(s)), this));
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException(e);
        }
    }

    // Pass Through ///////////////////////////////////////////////////////////

    public void setParent(Tree.Node p)    { wrapped.setParent(p); }
    public Tree.Node getParent()          { return wrapped.getParent(); }
    public XML.Attributes getAttributes() { return wrapped.getAttributes(); }
    public XML.Prefixes getPrefixes()     { return wrapped.getPrefixes(); }
    public List getChildren()             { return wrapped.getChildren(); }
    public String getQName()              { return wrapped.getQName(); }
    public String getLocalName()          { return wrapped.getLocalName(); }
    public String getPrefix()             { return wrapped.getPrefix(); }
    public String getUri()                { return wrapped.getUri(); }

    /** Works up the Element object model until an instance of a JSScope is found. */
    private static JSScope findScope(Tree.Node e) {
        while (e != null && !(e instanceof JSScope)) e = e.getParent();
        return (JSScope)e;
    }

    /** A JSElement with the element attributes merged with a second
     *  element.
     *
     *  All functions of the XML.Element interface are mapped onto the
     *  primary element, except <tt>getAttributes()</tt>. This function
     *  returns a MergedAttr instance with the <b>secondary</b> element
     *  acting as the primary attribute source.
     */
    public static class Merge extends JSElement {
        private final XML.Attributes a;
        public Merge(XML.Element wrapped, XML.Element merge) {
            super(wrapped);
            a = new MergeAttr(merge.getAttributes(), wrapped.getAttributes());
        }
        public XML.Attributes getAttributes() { return a; }
    }

    /** Creates a single view onto two sets of Attributes, first
     *  checking the <tt>primary</tt> array for an entry, or
     *  otherwise returning any matching entry in the
     *  <tt>secondary</tt> Attributes object.
     *
     *  FIXME: toXML() produces invalid XML if qname in both a and b.
     */
    public static final class MergeAttr implements XML.Attributes {
        private final XML.Attributes a, b;
        public MergeAttr(XML.Attributes primary, XML.Attributes secondary) {
            a = primary; b = secondary;
        }
        public int getIndex(String qname) {
            int i = a.getIndex(qname); if (i >= 0) return i;
                i = b.getIndex(qname); if (i >= 0) return i + a.attrSize();
            return -1;
        }
        public int getIndex(String uri, String key) {
            int i = a.getIndex(uri, key); if (i >= 0) return i;
                i = b.getIndex(uri, key); if (i >= 0) return i + b.attrSize();
            return -1;
        }
        public String getKey(int i) {
            return i >= a.attrSize() ? b.getKey(i-a.attrSize()) : a.getKey(i); }
        public String getVal(int i) {
            return i >= a.attrSize() ? b.getVal(i-a.attrSize()) : a.getVal(i); }
        public String getUri(int i) {
            return i >= a.attrSize() ? b.getUri(i-a.attrSize()) : a.getUri(i); }
        public String getPrefix(int i) {
            return i >= a.attrSize() ? b.getUri(i-a.attrSize()) : a.getUri(i); }
        public String getQName(int i) {
            return i >= a.attrSize() ? b.getUri(i-a.attrSize()) : a.getUri(i); }
        public int attrSize() { return a.attrSize() + b.attrSize(); }
    }
}
